<?php

namespace Drupal\entity_extender\Plugin\EntityExtender;

use Drupal\Core\Cache\CacheableDependencyInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Field\EntityReferenceFieldItemListInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Plugin\PluginBase;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\entity_extender\Component\Memo;
use Drupal\entity_extender\EntityExtenderManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * The base entity extender.
 */
abstract class EntityExtender extends PluginBase implements CacheableDependencyInterface {

  use Memo;

  /**
   * An instance of the entity.
   *
   * @var \Drupal\Core\Entity\EntityInterface
   */
  protected EntityInterface $entity;

  /**
   * A logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected LoggerChannelInterface $logger;

  /**
   * An instance of the "entity_extender.manager" service.
   *
   * @var \Drupal\entity_extender\EntityExtenderManagerInterface
   */
  protected EntityExtenderManagerInterface $entityExtenderManager;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    array $plugin,
    EntityInterface $entity,
    EntityExtenderManagerInterface $entity_extender_manager,
    TranslationInterface $string_translation,
    LoggerChannelFactoryInterface $logger_factory
  ) {
    \assert(!empty($plugin['id']));
    parent::__construct([], $plugin['id'], $plugin);

    $this->entity = $entity;
    $this->logger = $logger_factory->get(\sprintf('extender:%s:%s', $this->pluginId, $this->id()));
    $this->stringTranslation = $string_translation;
    $this->entityExtenderManager = $entity_extender_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    EntityExtenderManagerInterface $entity_extender_manager,
    EntityInterface $entity,
    array $plugin,
    ...$arguments
  ) {
    return new static(
      $plugin,
      $entity,
      $entity_extender_manager,
      $container->get('string_translation'),
      $container->get('logger.factory'),
      ...$arguments,
    );
  }

  /**
   * {@inheritdoc}
   */
  public function __toString(): string {
    return $this->label();
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheContexts(): array {
    return $this->entity->getCacheContexts();
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheMaxAge(): int {
    return $this->entity->getCacheMaxAge();
  }

  /**
   * {@inheritdoc}
   */
  public function getCacheTags(): array {
    return $this->entity->getCacheTags();
  }

  /**
   * {@inheritdoc}
   */
  public function id(): ?int {
    $id = $this->entity->id();

    return $id === NULL ? $id : (int) $id;
  }

  /**
   * {@inheritdoc}
   */
  public function uuid(): ?string {
    return $this->entity->uuid();
  }

  /**
   * {@inheritdoc}
   */
  public function bundle(): string {
    return $this->entity->bundle();
  }

  /**
   * {@inheritdoc}
   */
  public function label(): string {
    return (string) $this->entity->label();
  }

  /**
   * {@inheritdoc}
   */
  public function getEntity(): EntityInterface {
    return $this->entity;
  }

  /**
   * Returns the list of referenced entities within a field.
   *
   * @param string $field_name
   *   The name of the entity reference field.
   *
   * @return \Drupal\Core\Entity\EntityInterface[]
   *   The list of referenced entities within a field.
   *
   * @throws \InvalidArgumentException
   *   When the field name is invalid.
   */
  protected function getReferencedEntities(string $field_name): array {
    return static::useMemo(function () use ($field_name): array {
      $list = $this->getList($field_name);

      if ($list === NULL) {
        $this->logger->warning('The "@uuid" is not a fieldable entity so cannot have references.', [
          '@uuid' => $this->uuid(),
        ]);

        return [];
      }

      if (!$list instanceof EntityReferenceFieldItemListInterface) {
        $this->logger->warning('The "@field_name" is not a reference field of "@uuid" entity.', [
          '@field_name' => $field_name,
          '@uuid' => $this->uuid(),
        ]);

        return [];
      }

      return $list->referencedEntities();
    }, $this, __METHOD__, $field_name);
  }

  /**
   * Returns the items list of a field.
   *
   * @param string $field_name
   *   The name of a field.
   * @param string|null $constraint
   *   The name of a constraint to assert on a field.
   *
   * @return mixed
   *   The items list of a field.
   *
   * @throws \InvalidArgumentException
   *   When the field name is invalid.
   */
  protected function getList(string $field_name, string $constraint = NULL): ?FieldItemListInterface {
    // Do no check for "hasField" here because it becomes harder to understand:
    // why this method returned "NULL" - because of missing field or missing
    // value. In case field is missing exception will be thrown.
    if ($this->entity instanceof FieldableEntityInterface) {
      // Do not use "$field->getConstraint()" since there's an "isset()" inside
      // that doesn't work when a constraint has "NULL" as a configuration.
      \assert($constraint === NULL || \array_key_exists($constraint, $this->entity->getFieldDefinition($field_name)->getConstraints()));

      return $this->entity->get($field_name);
    }

    return NULL;
  }

  /**
   * Returns the value of a field.
   *
   * @param string $field_name
   *   The name of a field.
   * @param string|null $constraint
   *   The name of a constraint to assert on a field.
   *
   * @return mixed
   *   The value of a field.
   *
   * @throws \InvalidArgumentException
   *   When the field name is invalid.
   */
  protected function get(string $field_name, string $constraint = NULL) {
    return ($list = $this->getList($field_name, $constraint)) ? $list->value : NULL;
  }

}
