<?php

namespace Drupal\Tests\entity_extender\Unit;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\entity_extender\Component\Memo;
use Drupal\Tests\UnitTestCase;

/**
 * Tests memoization.
 *
 * @group entity_extender
 */
class MemoUnitTest extends UnitTestCase {

  /**
   * {@inheritdoc}
   */
  protected function getEntity(int $calls): EntityInterface {
    $entity = $this->createMock(EntityInterface::class);
    $entity
      ->expects(static::exactly($calls))
      ->method('uuid')
      ->willReturn('1f70ad30-5608-4012-b719-a71d2ad6ea0d');

    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  protected function getAccount(int $calls): AccountInterface {
    $account = $this->createMock(AccountInterface::class);
    $account
      ->expects(static::exactly($calls))
      ->method('getEmail')
      ->willReturn('test@example.com');

    return $account;
  }

  /**
   * Tests input arguments.
   */
  public function testInputs(): void {
    $method = new \ReflectionMethod(static::class, __FUNCTION__);

    foreach ([1, 2] as $actual) {
      static::assertSame(1, Memo::useMemo(static fn (): int => $actual, $method));
      // Add dependency as an input.
      static::assertSame($actual, Memo::useMemo(static fn (): int => $actual, $actual, $method));
    }

    static::assertSame(1, Memo::useMemo(static fn (): int => 1, $method));
    // Even despite the inputs are the same the call is on a new line so
    // it's considered as a new one.
    static::assertSame(2, Memo::useMemo(static fn (): int => 2, $method));

    // Two calls on the same line produce a single result when the same
    // inputs are used.
    static::assertSame(1, Memo::useMemo(static fn (): int => 1, $method)); static::assertSame(1, Memo::useMemo(static fn (): int => 2, $method));
  }

  /**
   * Tests invalid argument.
   */
  public function testInvalidInput(): void {
    $this->expectException(\InvalidArgumentException::class);
    $this->expectExceptionMessage('The "object" type is not supported.');

    static::assertSame(2, Memo::useMemo(static fn (): int => 2, new \stdClass()));
  }

  /**
   * Tests memoization.
   *
   * @param bool $use_entity
   *   The state of whether to use an entity as an argument.
   * @param bool $use_account
   *   The state of whether to use an account as an argument.
   * @param \Throwable|null $exception
   *   The expected exception to be thrown during the use of a static cache.
   *
   * @dataProvider provider
   */
  public function test(bool $use_entity, bool $use_account, ?\Throwable $exception): void {
    if ($exception !== NULL) {
      $this->expectException(\get_class($exception));
      $this->expectExceptionMessage($exception->getMessage());
    }

    $iterations_count = 5;
    $executions_count = 0;

    $entity = $use_entity ? $this->getEntity($iterations_count) : NULL;
    $account = $use_account ? $this->getAccount($iterations_count) : NULL;

    for ($iterations_passed = 0; $iterations_passed < $iterations_count; $iterations_passed++) {
      static::assertSame('bla', Memo::useMemo(static function () use (&$executions_count): string {
        $executions_count++;

        return 'bla';
      }, $entity, $account));
    }

    static::assertSame($iterations_count, $iterations_passed);
    static::assertSame(1, $executions_count);
  }

  /**
   * {@inheritdoc}
   */
  public function provider(): array {
    return [
      [FALSE, FALSE, new \AssertionError('assert(\count($pointer) > 2)')],
      [TRUE, FALSE, NULL],
      [FALSE, TRUE, NULL],
      [TRUE, TRUE, NULL],
    ];
  }

}
